

make(File) :- [File], make.

% just to avoid BProlog throwing the "predicate does not exist" error
group_switches(_,_) :- fail.

make :-
    tell('tune_params.fld'),
    make_header,
    findall(Sw-Vals,get_sw(Sw,Vals),List),
    ( findall(Name-Switches,group_switches(Name,Switches),GroupList), GroupList \= [] ->
        make_tabs_begin,
        build_tabbed_interface(GroupList),
        make_tabs_end
    ;
        compute_total_height(List,40,H),
        make_pack_begin(H),
        build_interface(List,110,70),
        make_pack_end
    ),
    make_footer,
    told,
    tell('tune_params_output.cxx'),
    output_header,
    output_code(List),
    output_footer,
    output_header1,
    output_code1(List),
    output_footer1,
    output_header2,
    output_code2(List),
    output_footer2,
    output_header3,
    output_code3(List),
    output_footer3,
    told.

slider_height(18).
slider_sep(0).

compute_total_height([],H,H).
compute_total_height([_-[_,Vals,_]|R],AH,H) :-
    length(Vals,N), slider_height(SH), slider_sep(S),
    NAH is AH + N*(SH+S) + 40,
    compute_total_height(R,NAH,H).

make_header :-
    writeln('# data file for the Fltk User Interface Designer (fluid)'),
    writeln('version 1.0109'),
    writeln('header_name {.h}'),
    writeln('code_name {.cxx}'),
    writeln('Function {make_window()} {open'),
    writeln('} {'),
    writeln('  Fl_Window param_tune_window {open'),
    writeln('    label {Parameter tuning}'),
    writeln('    xywh {0 0 640 480} type Double visible'),
    writeln('  } {'),
    writeln('   Fl_Button {} {'),
    writeln('     label {Load Default}'),
    writeln('     callback {load_default_compact();}'),
    writeln('     xywh {35 450 105 25} box GTK_UP_BOX'),
    writeln('   }'),
    writeln('   Fl_Button {} {'),
    writeln('     label {Load Parameters}'),
    writeln('     callback {load_parameters_compact();}'),
    writeln('     xywh {165 450 155 25} box GTK_UP_BOX'),
    writeln('   }'),
    writeln('   Fl_Button {} {'),
    writeln('     label {Save Parameters}'),
    writeln('     callback {normalize_parameters(); save_parameters(); save_parameters_compact();}'),
    writeln('     xywh {325 450 155 25} box GTK_UP_BOX'),
    writeln('   }'),
    writeln('   Fl_Button {} {'),
    writeln('     label {Done}'),
    writeln('     callback {param_done();}'),
    writeln('     xywh {545 450 85 25} box GTK_UP_BOX'),
    writeln('   }').
make_tabs_begin  :-
    writeln('    Fl_Tabs {} {'),
    writeln('      label {} open'),
    writeln('      xywh {5 5 630 440} box GTK_UP_BOX'),
    writeln('    } {').
make_pack_begin(H) :-
    writeln('     Fl_Scroll {} {open'),
    writeln('       xywh {15 45 610 380} type VERTICAL box NO_BOX'),
    writeln('     } {'),
    writeln('      Fl_Pack {} {open'),
      write('        xywh {20 50 570 '),write(H),writeln('}'),
    writeln('        code0 {o->spacing(15);}'),
    writeln('      } {').
make_pack_end :-
    writeln('      }'),
    writeln('     }').
make_tabs_end :-
    writeln('    }').
make_footer :-
    writeln('  }'),
    writeln('}').


compute_max_height([],M,M).
compute_max_height([_GroupName-Switches|Rest],CurrentMax,Max) :-
    collect_switches(Switches,List),
    compute_total_height(List,40,Height),
    NewMax is max(CurrentMax,Height),
    compute_max_height(Rest,NewMax,Max).


collect_switches([],[]).
collect_switches([Pattern|Rest],Result) :-
    findall(Pattern-Vals,get_sw(Pattern,Vals),List),
    collect_switches(Rest,RestList),
    append(List,RestList,Result).

build_tabbed_interface([]).
build_tabbed_interface([GroupName-Switches|Rest]) :-
    collect_switches(Switches,List),
    compute_total_height(List,40,Height),
    format('     Fl_Group {} {\n',[]),
    format('       label {~w} open\n',[GroupName]),
    format('       xywh {10 30 620 ~w} box NO_BOX align 0\n',[Height]),
    format('     } {\n',[]),
    make_pack_begin(Height),
    build_interface(List,110,70),
    make_pack_end,
    format('       }\n',[]),
    build_tabbed_interface(Rest).

build_interface([],_,_).
build_interface([Sw-Vals|Rest],X,Y) :-
    Vals = [unfixed,Values,Probs],
    flatten_switch_name(Sw,FlatSw),
    length(Values,N), slider_height(H), slider_sep(S),
    Height is N*(H+S)+35,
    format('        Fl_Group {} {\n',[]),
    format('          label {~w} open\n',[Sw]),
    format('          xywh {40 ~w 565 ~w} box NO_BOX labelfont 1 labelsize 20 labelcolor 8 align 21\n',[Y,Height]),
    format('        } {\n',[]),
    Y2 is Y+28,
    interface(FlatSw,[],Values,Probs,X,Y2),
    format('        }\n',[]),
    NextY is Y+20+Height,
    build_interface(Rest,X,NextY).


flatten_switch_name(Sw,FlatSw) :-
    Sw =.. [F|Args],
    make_safe_functor(F,SafeF),
    flatten_arguments(Args,SafeF,FlatSw).
flatten_arguments([],F,F).
flatten_arguments([A|Args],F,Flat) :-
    flatten_switch_name(A,FlatA),
    atom_concat(F,'_',F_),
    atom_concat(F_,FlatA,FlatB),
    flatten_arguments(Args,FlatB,Flat).

make_safe_functor(N,F) :- number(N), !, number_to_functor(N,F).
make_safe_functor(+,plus) :- !.
% probably will need to add other stuff that is not allowed in C variable names
make_safe_functor(F,F).
number_to_functor(N,F) :- integer(N),N>=0, term2atom(N,F).
number_to_functor(N,F) :- N<0, N1 is abs(N), number_to_functor(N1,PF), atom_concat(minus,PF,F).




interface(_,_,[],[],_,_).
interface(Sw,PrevValues,[NonFlatV|Values],[P|Probs],X,Y) :-
    flatten_switch_name(NonFlatV,V),
    format('         Fl_Value_Slider ~w_~w {\n',[Sw,V]),
    format('            label {~w}\n',[NonFlatV]),
    format('            callback {balance(~w_~w,',[Sw,V]),
    length(PrevValues,L1), length(Values,L2), L is L1+L2,
    write(L),
    value_list_output(PrevValues,Sw),
    value_list_output(Values,Sw),
    format(');}\n',[]),
    slider_height(H),
    format('            xywh {~w ~w 475 ~w} type {Horz Fill} box GTK_THIN_DOWN_BOX color 51 selection_color 72 align 4 value ~w\n',[X,Y,H,P]),
    format('         }\n',[]),
    slider_sep(S),
    NextY is Y + H + S,
    interface(Sw,[V|PrevValues],Values,Probs,X,NextY).

value_list_output([],_).
value_list_output([NonFlatV|Values],Sw) :- 
    flatten_switch_name(NonFlatV,V),
    format(',~w_~w',[Sw,V]), 
    value_list_output(Values,Sw).



output_header :-
    writeln('void save_parameters() {'),
    writeln('  std::ofstream file;'),
    writeln('  file.open("saved_parameters.pl");'),
    writeln('  file.precision(1000);'),
      write('  file').
output_footer :-
    writeln(';'),
    writeln('  file.close();'),
    writeln('}').
output_code([]).
output_code([Sw-Vals|Rest]) :-
    Vals = [unfixed,Values,_Probs],
    flatten_switch_name(Sw,FlatSw),
% only outputting the set_sw statements
%    format('\n  << "values(~w,~w).\\n"\n',[Sw,Values]),
    format('  << ":- set_sw(~w,["\n',[Sw]),
    output_code_probs(FlatSw,Values),
    format('  << "]).\\n"\n',[]),
    output_code(Rest).
output_code_probs(_,[]).
output_code_probs(Sw,[V|Values]) :-
    flatten_switch_name(V,FlatV),
    format('  \t<< ~w_~w->value()',[Sw,FlatV]),
    (Values = [_|_] ->
        format(' << ", "',[])
    ;   true),
    nl,
    output_code_probs(Sw,Values).


output_header1 :-
    writeln('void do_save_parameters_compact(std::ofstream* file) {'),
    writeln('  file->precision(1000);').
output_footer1 :-
%    writeln('  std::cout << "saved parameters\\n";'),
    writeln('  file->close();'),
    writeln('}').
output_code1([]).
output_code1([Sw-Vals|Rest]) :-
    Vals = [unfixed,Values,_Probs],
    flatten_switch_name(Sw,FlatSw),
    output_code_probs1(FlatSw,Values),
    output_code1(Rest).
output_code_probs1(_,[]).
output_code_probs1(Sw,[V|Values]) :-
    flatten_switch_name(V,FlatV),
    format(' *file << ~w_~w->value() << "\\n";\n',[Sw,FlatV]),
    output_code_probs1(Sw,Values).

output_header2 :-
    writeln('void do_load_parameters_compact(std::ifstream* file) {'),
    writeln('  double v;').
%    writeln('  std::cout << "loading parameters\\n";').
output_footer2 :-
%    writeln('  std::cout << "loaded parameters\\n";'),
    writeln('  file->close();'),
    writeln('}').
output_code2([]).
output_code2([Sw-Vals|Rest]) :-
    Vals = [unfixed,Values,_Probs],
    flatten_switch_name(Sw,FlatSw),
    output_code_probs2(FlatSw,Values),
    output_code2(Rest).
output_code_probs2(_,[]).
output_code_probs2(Sw,[V|Values]) :-
    flatten_switch_name(V,FlatV),
    format('  \t*file >> v;  ~w_~w->value(v);\n',[Sw,FlatV]),
    output_code_probs2(Sw,Values).

output_header3 :-
    writeln('void normalize_parameters() {').
output_footer3 :-
    writeln('}').
output_code3([]).
output_code3([Sw-Vals|Rest]) :-
    Vals = [unfixed,Values,_Probs],
    flatten_switch_name(Sw,FlatSw),
    length(Values,N),
    format('  normalize(~w',[N]),
    output_code_probs3(FlatSw,Values),
    format(');\n',[]),
    output_code3(Rest).
output_code_probs3(_,[]).
output_code_probs3(Sw,[V|Values]) :-
    flatten_switch_name(V,FlatV),
    format(', ~w_~w',[Sw,FlatV]),
    output_code_probs3(Sw,Values).
